#!/usr/bin/perl
#
#  exons -- read the file of gene/exon positions
#
#   Entry points:
#
#       get_exons 
#               For each exon specified in the file, get_exons creates a record
#               (implemented as parallel arrays) with a begin point, an end
#               point, and a transcription direction (1 = left-to-right,
#               -1 = right-to-left, 0 = not applicable).  For each gene, there
#               are two such records, one for the gene (distinguished from all
#               other records by having gene_name != NULL) followed immediately
#               by one for the CDS limits.  If the file contains no "+" line
#               for that gene, then the limits are both 0.
#
#       exon_header
#		Returns the header from an exons file if one exists, NULL otherwise.	
#

use 5.6.0;
#use strict;
#use warnings;


my (@beg, @end, @dir, @gene_name, @exon_name);		# Global parallel arrays.


# Read an exons file into the parallel arrays.

sub get_exons() {

    my $n = -1;						# Global array index.
    my ($beg, $end, $dir, $gene_name, $exon_name);	# Single elements to add to arrays.
    my $first_exon = 0;					# Index position of first exon.
    my $nerr = 0;					# Number of errors reported.
    my $head = 0;					# Have we read in a header yet?

    while (<EXON_FILE>) {
	next if /^\s*$/;
	next if /^\s*#/;
	unless ($head) {		# Skip header if there is one.
	    $head = 1;
	    next unless /^[>|<]/;
	}
        if (/^[>|<]/) {
            $n++;
            $dir = (/^</ ? -1 : /^>/ ? 1 : 0);
            ($beg, $end, $gene_name) = /^\s*[>|<]\s*(\d+)\s+(\d+)\s*(.*)/;
            $nerr = error($_, "Improper endpoints.", $nerr)
                if (!defined($beg) || !defined($end) || $beg <= 0 || $beg > $end);
            set_exon($n, $beg, $end, $dir, $gene_name || " ", "");
            $n++;               # CDS information.
            set_exon($n, 0, 0, $dir, "", "");
            $first_exon = $n+1;
        } elsif (/^\+/) {
            $n++;
            $nerr = error($_, "Expecting just two numbers.", $nerr)
                if (/^\s*\+\s*\d+\s+\d+\s+\d+/);
            ($beg, $end) = /^\s*\+\s*(\d+)\s+(\d+)/;
            $nerr = error($_, "Improper endpoints.", $nerr)
                if (!defined($beg) || !defined($end) || $beg <= 0 || $beg > $end);
            $nerr = error($_, "No gene named for CDS positions.", $nerr)
                if (!$first_exon);
            $nerr = error($_, "CDS positions outside of gene.", $nerr)
                if ($beg < $beg[$first_exon-2] || $end > $end[$first_exon-2]);
            $nerr = error($_, "Redundant CDS positions.", $nerr)
                if ($beg[$first_exon-1] != 0);
            set_exon ($first_exon - 1, $beg, $end, $dir, "", "");
            $n--;
        } else {
            $n++;
            ($beg, $end, $exon_name) = /^\s*(\d+)\s+(\d+)\s*(.*)/;
            $nerr = error($_, "Improper endpoints.", $nerr)
                if (!defined($beg) || !defined($end) || $beg <= 0 || $beg > $end);
            $nerr = error($_, "Exons out of order.", $nerr)
                if ($n > $first_exon && $beg <= $end[$n-1]);
            set_exon($n, $beg, $end, $dir, "", $exon_name);
        }
    }
    exit($nerr) if ($nerr);
    return ($n + 1, \@beg, \@end, \@dir, \@gene_name, \@exon_name);



    # Print error message for exons file and die.

    sub error {

        my ($line, $message, $nerrs) = @_;

        warn "\nError in exons file at line $.\n$line$message\n";
        return ++$nerrs;
    }


    # Add information to parallel arrays.

    sub set_exon {

        my ($n, $beg, $end, $dir, $gene_name, $exon_name) = @_;

	return if ($n < 0);
        $beg[$n] = $beg || 0;
        $end[$n] = $end || 0;
        $dir[$n] = $dir || 0;
        $gene_name[$n] = $gene_name || "";
        $exon_name[$n] = $exon_name || "";
    }

}


# Returns the header from an exons file if one exists, NULL otherwise.

sub exon_header {

    local *EXON_FILE = $_[0];
    seek(EXON_FILE, 0, 0);

    while (<EXON_FILE>) {
	next if (/^\s*$/ || /^\s*#/);
	return "" if /^\s*[>|<]/;
	return $_;
    }
    return;
}



return 1;
